import math
import random
import warnings
from typing import Union, List, Iterable

import torch
from torch import nn
from torch.nn import functional as F

from .tensor import SigmaReparamTensor, AnchoredReparamTensor, NormReparamTensor


class SpectralNormLinear(nn.Module):
    """
    Inspired by Apple's Spectral Normed Linear Layers
        (https://github.com/apple/ml-sigma-reparam)
    """

    def __init__(self, in_features: int, out_features: int, bias: bool = True):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.use_bias = bias

        self.weights = None

        # Define the bias vector as a learnable parameter if required.
        if self.use_bias:
            self.bias = nn.Parameter(torch.empty(out_features))
        else:
            # If no bias, register it as None.
            # This is important so that PyTorch doesn't complain when saving/loading the model.
            self.register_parameter("bias", None)

        self.reset_parameters()

    def reset_parameters(self) -> None:
        weights = torch.empty(self.out_features, self.in_features)
        stdv = 1.0 / math.sqrt(self.in_features)
        nn.init.uniform_(weights, a=-stdv, b=stdv)
        if self.use_bias:
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(weights)
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            nn.init.uniform_(self.bias, -bound, bound)
        self.weights = SigmaReparamTensor(weights)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return F.linear(x, self.weights(), self.bias)

    def __repr__(self) -> str:
        # Optional: A nice representation for printing the module.
        return (
            f"SpectralNormFeedForward(in_features={self.in_features},"
            f"out_features={self.out_features}, bias={self.use_bias})"
        )


class AnchoredLinear(nn.Module):
    """
    ...
    """

    def __init__(self, in_features: int, out_features: int, bias: bool = True):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.use_bias = bias

        self.weights = None

        # Define the bias vector as a learnable parameter if required.
        if self.use_bias:
            self.bias = nn.Parameter(torch.empty(out_features))
        else:
            # If no bias, register it as None.
            # This is important so that PyTorch doesn't complain when saving/loading the model.
            self.register_parameter("bias", None)

        self.reset_parameters()

    def reset_parameters(self) -> None:
        weights = torch.empty(self.out_features, self.in_features)
        stdv = 1.0 / math.sqrt(self.in_features)
        nn.init.uniform_(weights, a=-stdv, b=stdv)
        if self.use_bias:
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(weights)
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            nn.init.uniform_(self.bias, -bound, bound)
        self.weights = AnchoredReparamTensor(weights)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return F.linear(x, self.weights(), self.bias)

    def __repr__(self) -> str:
        # Optional: A nice representation for printing the module.
        return (
            f"AnchoredLinear(in_features={self.in_features},"
            f"out_features={self.out_features}, bias={self.use_bias})"
        )


class WeightNormedLinear(nn.Module):
    """
    ...
    """

    def __init__(self, in_features: int, out_features: int, bias: bool = True):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.use_bias = bias

        self.weights = None

        # Define the bias vector as a learnable parameter if required.
        if self.use_bias:
            self.bias = nn.Parameter(torch.empty(out_features))
        else:
            # If no bias, register it as None.
            # This is important so that PyTorch doesn't complain when saving/loading the model.
            self.register_parameter("bias", None)

        self.reset_parameters()

    def reset_parameters(self) -> None:
        weights = torch.empty(self.out_features, self.in_features)
        stdv = 1.0 / math.sqrt(self.in_features)
        nn.init.uniform_(weights, a=-stdv, b=stdv)
        if self.use_bias:
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(weights)
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            nn.init.uniform_(self.bias, -bound, bound)
        self.weights = NormReparamTensor(weights)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return F.linear(x, self.weights(), self.bias)

    def __repr__(self) -> str:
        return (
            f"WeightNormedLinear(in_features={self.in_features},"
            f"out_features={self.out_features}, bias={self.use_bias})"
        )


class RecyclingLinear(nn.Module):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = True,
        row_recycling_rate: float = 0.0,
        column_recycling_rate: float = 0.0,
        adaptive=False,
        xglu=False,
    ):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.bias = bias
        self.xglu = xglu
        self.linear = nn.Linear(in_features, out_features, bias=bias)
        self.row_recycling_rate = row_recycling_rate
        self.column_recycling_rate = column_recycling_rate
        self.adaptive = adaptive
        self.optimisers = []
        self.initial_learning_rates = []
        self._warned_about_registration = False

    def register_optimiser(self, optimiser: torch.optim.Optimizer):
        self.optimisers.append(optimiser)
        self.initial_learning_rates.append(self._get_learning_rate(optimiser))
        if self.initial_learning_rates[-1] == 0.0:
            warnings.warn(
                "Learning rate of registered optimiser was 0.0 - make sure "
                "you haven't initialised a scheduler before registering the "
                "optimiser",
                stacklevel=2,
            )

    def _get_learning_rate(self, optimiser: torch.optim.Optimizer):
        for group in optimiser.param_groups:
            for param in group["params"]:
                if param is self.linear.weight:
                    return group["lr"]

    def _get_multiplier(self):
        if not self.adaptive or not self.optimisers:
            return 1.0
        else:
            init = self.initial_learning_rates
            current = [self._get_learning_rate(o) for o in self.optimisers]
            pairs = zip(current, init, strict=True)
            multipliers = [a / b for a, b in pairs if b != 0.0]
            return min(multipliers) if multipliers else 0.0

    def reset_rows(self, indices, incoming_data=None):
        """
        Resets rows.
        If incoming_data is provided, resets to the centroid (mean) of that data.
        If not, resets to the mean of existing weights.
        """
        if not torch.is_tensor(indices):
            idx_tensor = torch.as_tensor(
                list(indices), dtype=torch.long, device=self.linear.weight.device
            )
        else:
            idx_tensor = indices

        if idx_tensor.numel() == 0:
            return

        if incoming_data is not None:
            target_center = self._mean_input_weights(incoming_data)
        else:
            target_center = self._mean_value_weights()

        target_center = target_center.expand(idx_tensor.size(0), -1)

        if self.xglu:
            gate_indices = idx_tensor
            value_indices = idx_tensor + (self.linear.out_features // 2)
            self._update_weights(gate_indices, 0, target_center, self.optimisers)
            self._update_weights(value_indices, 0, target_center, self.optimisers)
        else:
            self._update_weights(idx_tensor, 0, target_center, self.optimisers)

    def reset_columns(self, indices):
        if not torch.is_tensor(indices):
            idx_tensor = torch.as_tensor(
                list(indices), dtype=torch.long, device=self.linear.weight.device
            )
        else:
            idx_tensor = indices

        if idx_tensor.size(0):
            random_weights = self._random_weights(
                self.linear.weight.size(0), indices.size(0)
            )
            # Make random col weights quiet so they don't introduce loud noise...
            # ...but not so quiet that FP16 zeros them and ruins symmetry breaking!
            random_weights *= 0.1
            self._update_weights(indices, 1, random_weights, self.optimisers)  # dim
        else:
            return

    def forward(self, x):
        if self.training and self.optimisers:
            self.reset_rows(self.get_reset_indices(0))
            self.reset_columns(self.get_reset_indices(1))
        elif self.training and not self._warned_about_registration:
            warnings.warn(
                "RecyclingLinear: No optimiser registered. Recycling disabled.",
                stacklevel=2,
            )
            self._warned_about_registration = True

        return self.linear(x)

    def get_reset_indices(self, dim):
        base_rate = self.row_recycling_rate if dim == 0 else self.column_recycling_rate
        p = base_rate * self._get_multiplier()
        if dim == 0:
            if self.xglu:
                sample_space = self.linear.out_features // 2
            else:
                sample_space = self.linear.out_features
        elif dim == 1:
            sample_space = self.linear.in_features
        else:
            raise ValueError("`dim` must be 0 or 1")

        # Sample the indices
        probs = torch.rand(sample_space, device=self.linear.weight.device)
        mask = probs < p
        if mask.any():
            return torch.nonzero(mask).squeeze(-1)
        else:
            return torch.tensor([], dtype=torch.long, device=self.linear.weight.device)

    def _random_weights(self, rows, columns):
        device = self.linear.weight.device
        weights = self.linear.weight.data
        stdv = 1.0 / math.sqrt(weights.size(1))
        random_weights = torch.rand(rows, columns, device=device)
        random_weights -= 0.5  # Range [-0.5, +0.5]
        random_weights *= 2.0 * stdv  # Range [-stdv, +stdv]
        return random_weights

    def _mean_input_weights(self, input):
        reduce_dims = list(range(input.ndim - 1))
        data_mean = input.detach().mean(dim=reduce_dims, keepdim=True)

        weights = self.linear.weight.data
        stdv = 1.0 / math.sqrt(weights.size(1))
        data_norm = data_mean.std() + 1e-6
        scale_factor = stdv / data_norm

        return data_mean * scale_factor

    def _mean_value_weights(self):
        """
        Only used when self.xglu
        """
        weights = self.linear.weight.data
        rows = weights.size(0)
        if self.xglu:
            return self.linear.weight[int(rows / 2) :].data.mean(dim=0, keepdim=True)
        else:
            return self.linear.weight.data.mean(dim=0, keepdim=True)

    def _mean_gate_weights(self):
        """
        Only used when self.xglu
        """
        weights = self.linear.weight.data
        rows = weights.size(0)
        return self.linear.weight[: int(rows / 2)].data.mean(dim=0, keepdim=True)

    def _update_weights(
        self,
        indices: Iterable[int],
        dim: int,
        data: torch.Tensor,
        optimisers: Union[
            List[torch.optim.Optimizer], torch.optim.Optimizer, None
        ] = None,
    ):
        if optimisers is None:
            optimisers = []
        if not isinstance(optimisers, list):
            optimisers = [optimisers]

        if not torch.is_tensor(indices):
            idx_tensor = torch.as_tensor(
                list(indices), dtype=torch.long, device=self.linear.weight.device
            )
        else:
            idx_tensor = indices

        if idx_tensor.numel() == 0:
            return

        with torch.no_grad():
            if dim == 0:
                self.linear.weight.data[idx_tensor] = data
            elif dim == 1:
                self.linear.weight.data[:, idx_tensor] = data
            else:
                raise ValueError("`dim` must be 0 or 1")

            self._reset_optim_state(self.linear.weight, idx_tensor, optimisers, dim=dim)

    def _reset_optim_state(self, param, idx_tensor, optimisers, dim):
        """
        Zeroes out the optimizer state for the given indices in a single operation.
        """
        for optimiser in optimisers:
            if param not in optimiser.state:
                continue
            state = optimiser.state[param]

            for _, buffer in state.items():
                if torch.is_tensor(buffer) and buffer.shape == param.shape:
                    # Vectorized zeroing
                    if dim == 0:
                        buffer[idx_tensor] = 0.0
                    else:
                        buffer[:, idx_tensor] = 0.0



import torch
from torch import nn
from torch.nn import functional as F


class ReLU(nn.Module):
    """
    A ReLU activation function with optional clamp and leakiness.
    """

    def __init__(
        self, clamp=True, leaky=True, negative_slope=0.01, clamp_max=6.0
    ) -> None:
        super().__init__()
        self.clamp = clamp
        self.leaky = leaky
        self.negative_slope = negative_slope
        self.clamp_max = clamp_max

    def forward(self, x):
        if self.leaky:
            relu = F.leaky_relu(x, negative_slope=self.negative_slope)
        else:
            relu = F.relu(x)
        if self.clamp:
            relu = torch.clamp(relu, max=self.clamp_max)
        return relu


class GELU(nn.Module):
    """
    A GELU activation function with optional clamp.
    """

    def __init__(self, clamp=True) -> None:
        super().__init__()
        self.clamp = clamp
        self.gelu = nn.GELU()

    def forward(self, x):
        gelu = self.gelu(x)
        if self.clamp:
            gelu = torch.clamp(gelu, max=6)
        return gelu


class Swish(nn.Module):
    """
    Implementation of (beta) Swish
    """

    def __init__(self) -> None:
        super().__init__()
        # Learnable parameter is called "swiglu beta" so that it is easy to find
        #   and exclude from weight decay
        self.swish_beta = nn.Parameter(torch.tensor([1.0]))

    def forward(self, x):
        return x * F.sigmoid(self.swish_beta * x)


class SquaredReLU(nn.Module):
    """
    Squared ReLU, as shown in "ReLU^2 wins" (https://arxiv.org/abs/2402.03804) to
      be as effective as SwiGLU for training LLMs, possibly because it can allow a
      NN to learn multyiplication, as noted by
      https://azizbelaweid.substack.com/p/what-is-swiglu-how-to-implement-it
    """

    def __init__(
        self, clamp=True, leaky=True, negative_slope: float = 0.01, clamp_max=6
    ) -> None:
        super().__init__()
        self.clamp = clamp
        self.leaky = leaky
        self.negative_slope = negative_slope
        self.clamp_max = clamp_max

    def forward(self, x):
        if self.leaky:
            relu = F.leaky_relu(x, negative_slope=self.negative_slope)
        else:
            relu = F.relu(x)
        relu_squared = relu**2
        if self.clamp:
            relu_squared = torch.clamp(relu_squared, max=self.clamp_max)
        return relu_squared


class XGLU(nn.Module):
    """
    Generic Gated Linear Unit
    """

    def __init__(self, activation_module: nn.Module) -> None:
        super().__init__()
        self.activation = activation_module

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        gate, value = x.chunk(2, dim=-1)
        return self.activation(gate) * value


def SquaredReGLU(clamp=True, leaky=True, negative_slope=0.01, clamp_max=6.0) -> XGLU:
    """
    Factory function that creates a GLU with a SquaredReLU activation.
    """
    activation_module = SquaredReLU(
        clamp=clamp, leaky=leaky, negative_slope=negative_slope, clamp_max=clamp_max
    )
    return XGLU(activation_module)


def SwiGLU() -> XGLU:
    """
    Factory function that creates a GLU with a Swish activation.
    """
    return XGLU(Swish())



import torch.nn as nn
import torch.nn.functional as F


class PadTensor(nn.Module):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.args = args
        self.kwargs = kwargs

    def forward(self, x):
        if sum(self.args[0]) == 0:
            return x
        else:
            return F.pad(x, *self.args, **self.kwargs)



"""
This code is from https://github.com/lucidrains/rotary-embedding-torch

It is provided along with the following copyright notice and license:

MIT License

Copyright (c) 2021 Phil Wang

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations
from math import pi

import torch

from torch.nn import Module
from torch import nn, einsum, broadcast_tensors, is_tensor, tensor, Tensor

# Gracefully find the best way to import autocast
try:
    from torch.amp import autocast as autocast_factory
except ImportError:
    # Fallback: For PyTorch 1.6 to 1.9
    from torch.cuda.amp import autocast

    def autocast_factory(_, enabled=True):
        """
        A wrapper that mimics the modern autocast signature but calls the older
        torch.cuda.amp.autocast, ignoring the device_type argument.
        """
        return autocast(enabled=enabled)


from einops import rearrange, repeat

from typing import Literal

# helper functions


def exists(val):
    return val is not None


def default(val, d):
    return val if exists(val) else d


# broadcat, as tortoise-tts was using it


def broadcat(tensors, dim=-1):
    broadcasted_tensors = broadcast_tensors(*tensors)
    return torch.cat(broadcasted_tensors, dim=dim)


def slice_at_dim(t, dim_slice: slice, *, dim):
    dim += t.ndim if dim < 0 else 0
    colons = [slice(None)] * t.ndim
    colons[dim] = dim_slice
    return t[tuple(colons)]


# rotary embedding helper functions


def rotate_half(x):
    x = rearrange(x, "... (d r) -> ... d r", r=2)
    x1, x2 = x.unbind(dim=-1)
    x = torch.stack((-x2, x1), dim=-1)
    return rearrange(x, "... d r -> ... (d r)")


@autocast_factory("cuda", enabled=False)
def apply_rotary_emb(
    freqs, t, start_index=0, scale=1.0, seq_dim=-2, freqs_seq_dim=None
):
    dtype = t.dtype

    if not exists(freqs_seq_dim):
        if freqs.ndim == 2 or t.ndim == 3:
            freqs_seq_dim = 0

    if t.ndim == 3 or exists(freqs_seq_dim):
        seq_len = t.shape[seq_dim]
        freqs = slice_at_dim(freqs, slice(-seq_len, None), dim=freqs_seq_dim)

    rot_dim = freqs.shape[-1]
    end_index = start_index + rot_dim

    assert rot_dim <= t.shape[-1], (
        f"feature dimension {t.shape[-1]} is not of sufficient size "
        "to rotate in all the positions {rot_dim}"
    )

    # Split t into three parts: left, middle (to be transformed), and right
    t_left = t[..., :start_index]
    t_middle = t[..., start_index:end_index]
    t_right = t[..., end_index:]

    # Apply rotary embeddings without modifying t in place
    t_transformed = (t_middle * freqs.cos() * scale) + (
        rotate_half(t_middle) * freqs.sin() * scale
    )

    out = torch.cat((t_left, t_transformed, t_right), dim=-1)

    return out.type(dtype)


# learned rotation helpers


def apply_learned_rotations(rotations, t, start_index=0, freq_ranges=None):
    if exists(freq_ranges):
        rotations = einsum("..., f -> ... f", rotations, freq_ranges)
        rotations = rearrange(rotations, "... r f -> ... (r f)")

    rotations = repeat(rotations, "... n -> ... (n r)", r=2)
    return apply_rotary_emb(rotations, t, start_index=start_index)


# classes


class RotaryEmbedding(Module):
    def __init__(
        self,
        dim,
        custom_freqs: Tensor | None = None,
        freqs_for: Literal["lang", "pixel", "constant"] = "lang",
        theta=10000,
        max_freq=10,
        num_freqs=1,
        learned_freq=False,
        use_xpos=False,
        xpos_scale_base=512,
        interpolate_factor=1.0,
        theta_rescale_factor=1.0,
        seq_before_head_dim=False,
        cache_if_possible=True,
        cache_max_seq_len=8192,
    ):
        super().__init__()
        # proposed by reddit user bloc97,
        # to rescale rotary embeddings to longer sequence length without fine-tuning
        # has some connection to NTK literature
        # https://www.reddit.com/r/LocalLLaMA/comments/14lz7j5/ntkaware_scaled_rope_allows_llama_models_to_have/

        theta *= theta_rescale_factor ** (dim / (dim - 2))

        self.freqs_for = freqs_for

        if exists(custom_freqs):
            freqs = custom_freqs
        elif freqs_for == "lang":
            freqs = 1.0 / (
                theta ** (torch.arange(0, dim, 2)[: (dim // 2)].float() / dim)
            )
        elif freqs_for == "pixel":
            freqs = torch.linspace(1.0, max_freq / 2, dim // 2) * pi
        elif freqs_for == "constant":
            freqs = torch.ones(num_freqs).float()

        self.cache_if_possible = cache_if_possible
        self.cache_max_seq_len = cache_max_seq_len

        self.register_buffer(
            "cached_freqs", torch.zeros(cache_max_seq_len, dim), persistent=False
        )
        self.cached_freqs_seq_len = 0

        self.freqs = nn.Parameter(freqs, requires_grad=learned_freq)

        self.learned_freq = learned_freq

        # dummy for device

        self.register_buffer("dummy", torch.tensor(0), persistent=False)

        # default sequence dimension

        self.seq_before_head_dim = seq_before_head_dim
        self.default_seq_dim = -3 if seq_before_head_dim else -2

        # interpolation factors

        assert interpolate_factor >= 1.0
        self.interpolate_factor = interpolate_factor

        # xpos

        self.use_xpos = use_xpos

        if not use_xpos:
            return

        scale = (torch.arange(0, dim, 2) + 0.4 * dim) / (1.4 * dim)
        self.scale_base = xpos_scale_base

        self.register_buffer("scale", scale, persistent=False)
        self.register_buffer(
            "cached_scales", torch.zeros(cache_max_seq_len, dim), persistent=False
        )
        self.cached_scales_seq_len = 0

        # add apply_rotary_emb as static method

        self.apply_rotary_emb = staticmethod(apply_rotary_emb)

    @property
    def device(self):
        return self.dummy.device

    def get_seq_pos(self, seq_len, device=None, dtype=None, offset=0):
        device = default(device, self.device)
        dtype = default(dtype, self.cached_freqs.dtype)

        return (
            torch.arange(seq_len, device=device, dtype=dtype) + offset
        ) / self.interpolate_factor

    def rotate_queries_or_keys(self, t, seq_dim=None, offset=0, scale=None):
        seq_dim = default(seq_dim, self.default_seq_dim)

        assert not self.use_xpos or exists(scale), (
            "you must use `.rotate_queries_and_keys` method instead and pass "
            "in both queries and keys, for length extrapolatable rotary embeddings"
        )

        device, dtype, seq_len = t.device, t.dtype, t.shape[seq_dim]

        seq = self.get_seq_pos(seq_len, device=device, dtype=dtype, offset=offset)

        freqs = self.forward(seq, seq_len=seq_len, offset=offset)

        if seq_dim == -3:
            freqs = rearrange(freqs, "n d -> n 1 d")

        return apply_rotary_emb(freqs, t, scale=default(scale, 1.0), seq_dim=seq_dim)

    def rotate_queries_with_cached_keys(self, q, k, seq_dim=None, offset=0):
        dtype, device, seq_dim = (
            q.dtype,
            q.device,
            default(seq_dim, self.default_seq_dim),
        )

        q_len, k_len = q.shape[seq_dim], k.shape[seq_dim]
        assert q_len <= k_len

        q_scale = k_scale = 1.0

        if self.use_xpos:
            seq = self.get_seq_pos(k_len, dtype=dtype, device=device)

            q_scale = self.get_scale(seq[-q_len:]).type(dtype)
            k_scale = self.get_scale(seq).type(dtype)

        rotated_q = self.rotate_queries_or_keys(
            q, seq_dim=seq_dim, scale=q_scale, offset=k_len - q_len + offset
        )
        rotated_k = self.rotate_queries_or_keys(k, seq_dim=seq_dim, scale=k_scale**-1)

        rotated_q = rotated_q.type(q.dtype)
        rotated_k = rotated_k.type(k.dtype)

        return rotated_q, rotated_k

    def rotate_queries_and_keys(self, q, k, seq_dim=None):
        seq_dim = default(seq_dim, self.default_seq_dim)

        assert self.use_xpos
        device, dtype, seq_len = q.device, q.dtype, q.shape[seq_dim]

        seq = self.get_seq_pos(seq_len, dtype=dtype, device=device)

        freqs = self.forward(seq, seq_len=seq_len)
        scale = self.get_scale(seq, seq_len=seq_len).to(dtype)

        if seq_dim == -3:
            freqs = rearrange(freqs, "n d -> n 1 d")
            scale = rearrange(scale, "n d -> n 1 d")

        rotated_q = apply_rotary_emb(freqs, q, scale=scale, seq_dim=seq_dim)
        rotated_k = apply_rotary_emb(freqs, k, scale=scale**-1, seq_dim=seq_dim)

        rotated_q = rotated_q.type(q.dtype)
        rotated_k = rotated_k.type(k.dtype)

        return rotated_q, rotated_k

    def get_scale(self, t: Tensor, seq_len: int | None = None, offset=0):
        assert self.use_xpos

        should_cache = (
            self.cache_if_possible
            and exists(seq_len)
            and (offset + seq_len) <= self.cache_max_seq_len
        )

        if (
            should_cache
            and exists(self.cached_scales)
            and (seq_len + offset) <= self.cached_scales_seq_len
        ):
            return self.cached_scales[offset : (offset + seq_len)]

        scale = 1.0
        if self.use_xpos:
            power = (t - len(t) // 2) / self.scale_base
            scale = self.scale ** rearrange(power, "n -> n 1")
            scale = repeat(scale, "n d -> n (d r)", r=2)

        if should_cache and offset == 0:
            self.cached_scales[:seq_len] = scale.detach()
            self.cached_scales_seq_len = seq_len

        return scale

    def get_axial_freqs(
        self, *dims, offsets: tuple[int | float, ...] | Tensor | None = None
    ):
        Colon = slice(None)
        all_freqs = []

        # handle offset

        if exists(offsets):
            if not is_tensor(offsets):
                offsets = tensor(offsets)

            assert len(offsets) == len(dims)

        # get frequencies for each axis

        for ind, dim in enumerate(dims):

            offset = 0
            if exists(offsets):
                offset = offsets[ind]

            if self.freqs_for == "pixel":
                pos = torch.linspace(-1, 1, steps=dim, device=self.device)
            else:
                pos = torch.arange(dim, device=self.device)

            pos = pos + offset

            freqs = self.forward(pos, seq_len=dim)

            all_axis = [None] * len(dims)
            all_axis[ind] = Colon

            new_axis_slice = (Ellipsis, *all_axis, Colon)
            all_freqs.append(freqs[new_axis_slice])

        # concat all freqs

        all_freqs = broadcast_tensors(*all_freqs)
        return torch.cat(all_freqs, dim=-1)

    @autocast_factory("cuda", enabled=False)
    def forward(self, t: Tensor, seq_len: int | None = None, offset=0):
        should_cache = (
            self.cache_if_possible
            and not self.learned_freq
            and exists(seq_len)
            and self.freqs_for != "pixel"
            and (offset + seq_len) <= self.cache_max_seq_len
        )

        if (
            should_cache
            and exists(self.cached_freqs)
            and (offset + seq_len) <= self.cached_freqs_seq_len
        ):
            return self.cached_freqs[offset : (offset + seq_len)].detach()

        freqs = self.freqs

        freqs = einsum("..., f -> ... f", t.type(freqs.dtype), freqs)
        freqs = repeat(freqs, "... n -> ... (n r)", r=2)

        if should_cache and offset == 0:
            self.cached_freqs[:seq_len] = freqs.detach()
            self.cached_freqs_seq_len = seq_len

        return freqs



import warnings
import math
from typing import Optional, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.checkpoint import checkpoint

from einops import rearrange

from .rope import RotaryEmbedding, apply_rotary_emb

try:
    from flash_attn import flash_attn_func

    print("Using flash-attn.")
    FLASH_ATTN = True
except ImportError:
    pass
    FLASH_ATTN = False


def scale_parameters(torch_module: nn.Module, factor: float):
    with torch.no_grad():
        for param in torch_module.parameters():
            param.mul_(factor)


def drop_path(
    x, drop_prob: float = 0.0, training: bool = False, scale_by_keep: bool = True
):
    """
    From https://github.com/huggingface/pytorch-image-models/blob/main/timm/layers/drop.py
    Copyright 2019 Ross Wightman
    See documentation and licence there.
    """
    if drop_prob == 0.0 or not training:
        return x
    keep_prob = 1 - drop_prob
    shape = (x.shape[0],) + (1,) * (
        x.ndim - 1
    )  # work with diff dim tensors, not just 2D ConvNets
    random_tensor = x.new_empty(shape).bernoulli_(keep_prob)
    if keep_prob > 0.0 and scale_by_keep:
        random_tensor.div_(keep_prob)
    return x * random_tensor


class DropPath(nn.Module):
    """
    From https://github.com/huggingface/pytorch-image-models/blob/main/timm/layers/drop.py
    Copyright 2019 Ross Wightman
    See documentation and licence there.
    """

    def __init__(self, drop_prob: float = 0.0, scale_by_keep: bool = True):
        super(DropPath, self).__init__()
        self.drop_prob = drop_prob
        self.scale_by_keep = scale_by_keep

    def forward(self, x):
        return drop_path(x, self.drop_prob, self.training, self.scale_by_keep)

    def extra_repr(self):
        return f"drop_prob={round(self.drop_prob, 3):0.3f}"


class MHAttention(nn.Module):
    """
    Multi-head self-attention using einops and optionally a custom linear layer.

    Forward method assumes q, k and v have the same embedding size and k and v
        are the same shape.

    Assumes bias=False and batch_first=True, as God intended.
    """

    def __init__(
        self,
        embed_dim,
        n_heads,
        dropout=0.0,
        causal=False,
        seq_len=None,
        linear_module: nn.Module = nn.Linear,
        utility_tokens=0,
        talking_heads=False,
        rotary_embedding=None,
        source_size=None,
        scaling="d",
        beta=1.0,
    ):
        """
        Args:
            scaling: how should the attention logits be scaled? Can be "sqrtd"
                to mimic the original Attention is All You Need approach of
                dividing by the sqrt of the embedding Dimension or "d" per
                "Tensor Programs V...". Default "d"
        """
        super().__init__()

        if rotary_embedding is not None:
            assert source_size is not None
        if causal:
            assert seq_len is not None

        self.talking_heads = talking_heads

        if self.talking_heads:
            self.head_projection = nn.Linear(n_heads, n_heads, bias=False)
            self.sample_projection = nn.Linear(n_heads, n_heads, bias=False)
        else:
            self.head_projection = None
            self.sample_projection = None

        self.embed_dim = embed_dim
        self.n_heads = n_heads
        assert embed_dim % n_heads == 0
        self.scaling = scaling
        self.beta = beta

        self.head_dim = self.embed_dim // self.n_heads

        if self.scaling == "sqrtd":
            self.scaling_factor = 1 / math.sqrt(self.head_dim)
        elif self.scaling == "d":
            # 8/d_model for backwards compatibility,
            #     per https://github.com/microsoft/mup
            self.scaling_factor = 8 / self.head_dim
        else:
            raise ValueError('`scaling` argument to MHAttention must be "d" or "sqrtd"')

        self.q_proj = linear_module(self.embed_dim, self.embed_dim, bias=False)
        self.k_proj = linear_module(self.embed_dim, self.embed_dim, bias=False)
        self.v_proj = linear_module(self.embed_dim, self.embed_dim, bias=False)

        self.out_proj = linear_module(self.embed_dim, self.embed_dim, bias=False)

        self.causal = causal
        self.seq_len = seq_len
        self.dropout = nn.Dropout(dropout)
        if self.causal:
            self.register_buffer(
                "mask",
                (torch.triu(torch.ones(seq_len, seq_len), diagonal=1) == 1)
                .unsqueeze(0)
                .unsqueeze(0),
            )
        self.rotary_embedding = rotary_embedding
        self.source_size = source_size
        self.utility_tokens = utility_tokens

        self.reset_parameters()

    @property
    def _kv_distance(self) -> float:
        """
        Calculates the cosine distance between the weight tensors of `self.k_proj`
            and `self.v_proj`.

        The cosine distance is defined as 1 - cosine_similarity (i.e. a value
            closer to 0 indicates higher similarity.
        """

        similarity = F.cosine_similarity(
            self.k_proj.weight.detach().flatten(),
            self.v_proj.weight.detach().flatten(),
            dim=0,
            eps=1e-8,
        ).item()

        return 1 - similarity

    def add_axial_rope(
        self, q: torch.Tensor, k: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Apply Axial RoPE to all tokens except utility tokens
        """

        if len(self.source_size) == 1:
            spatial_dimension_names = "D1"
            spatial_dimension_values = {"D1": self.source_size[0]}
        elif len(self.source_size) == 2:
            spatial_dimension_names = "D1 D2"
            spatial_dimension_values = {
                "D1": self.source_size[0],
                "D2": self.source_size[1],
            }
        elif len(self.source_size) == 3:
            spatial_dimension_names = "D1 D2 D3"
            spatial_dimension_values = {
                "D1": self.source_size[0],
                "D2": self.source_size[1],
                "D3": self.source_size[2],
            }
        else:
            raise NotImplementedError(
                "`source_size` must be a tuple of 1, 2 or 3 integers"
            )

        q = rearrange(q, "b t (h d) -> b h t d", h=self.n_heads)
        k = rearrange(k, "b t (h d) -> b h t d", h=self.n_heads)

        q_util, q_img = (
            q[:, :, : self.utility_tokens, :],
            q[:, :, self.utility_tokens :, :],
        )
        k_util, k_img = (
            k[:, :, : self.utility_tokens, :],
            k[:, :, self.utility_tokens :, :],
        )

        q_img = rearrange(
            q_img,
            f"b h ({spatial_dimension_names}) d -> b h {spatial_dimension_names} d",
            **spatial_dimension_values,
        )
        k_img = rearrange(
            k_img,
            f"b h ({spatial_dimension_names}) d -> b h {spatial_dimension_names} d",
            **spatial_dimension_values,
        )

        freqs = self.rotary_embedding.get_axial_freqs(*self.source_size)

        # norm Qs/Ks to protect axial rope, like https://arxiv.org/abs/2302.05442
        q_img = apply_rotary_emb(freqs, q_img)
        k_img = apply_rotary_emb(freqs, k_img)

        q_img = rearrange(
            q_img,
            f"b h {spatial_dimension_names} d -> b h ({spatial_dimension_names}) d",
        )
        k_img = rearrange(
            k_img,
            f"b h {spatial_dimension_names} d -> b h ({spatial_dimension_names}) d",
        )

        # Re-combine the utility tokens and the RoPE-enhanced sequence tokens
        q = torch.cat([q_util, q_img], dim=2)
        k = torch.cat([k_util, k_img], dim=2)

        q = rearrange(q, "b h t d -> b t (h d)")
        k = rearrange(k, "b h t d -> b t (h d)")

        return q, k

    def project_qkv(
        self, q: torch.Tensor, k: torch.Tensor, v: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        query_batch_size, query_tokens, query_features = q.size()
        key_batch_size, key_tokens, key_features = k.size()

        assert k.size() == v.size()
        assert query_features == key_features
        assert (
            (query_batch_size == key_batch_size)  # batch sizes are the same...
            or query_batch_size == 1  # ... or query is broadcastable
        )

        if self.causal:
            assert query_tokens == key_tokens
            assert query_tokens == self.seq_len

        q, k, v = self.q_proj(q), self.k_proj(k), self.v_proj(v)

        if self.rotary_embedding is not None:
            q, k = self.add_axial_rope(q, k)

        return q, k, v

    def forward(self, q, k, v):

        q, k, v = self.project_qkv(q, k, v)

        if FLASH_ATTN and not self.talking_heads:
            # Divide Q/K/V into heads
            q = rearrange(q, "b t (h d) -> b t h d", h=self.n_heads)
            k = rearrange(k, "b t (h d) -> b t h d", h=self.n_heads)
            v = rearrange(v, "b t (h d) -> b t h d", h=self.n_heads)

            output_with_heads = flash_attn_func(
                q,
                k,
                v,
                dropout_p=self.dropout.p if self.training else 0.0,
                softmax_scale=self.scaling_factor,
                causal=self.causal,
            )

            output_without_heads = rearrange(output_with_heads, "b t h d -> b t (h d)")

            return self.out_proj(output_without_heads)
        else:
            # Divide Q/K/V into heads
            q = rearrange(q, "b t (h d) -> b h t d", h=self.n_heads)
            k = rearrange(k, "b t (h d) -> b h t d", h=self.n_heads)
            v = rearrange(v, "b t (h d) -> b h t d", h=self.n_heads)

            qk_scores = q @ k.transpose(-1, -2)

            qk_scores *= self.scaling_factor

            if self.talking_heads:
                qk_scores = torch.einsum(
                    "b h i j, o h -> b o i j", qk_scores, self.head_projection.weight
                )

            # Apply mask if causal (must come before softmax)
            if self.causal:
                qk_scores.masked_fill_(self.mask, float("-inf"))

            qk_scores = F.softmax(qk_scores, dim=-1)

            if self.talking_heads:
                qk_scores = torch.einsum(
                    "b h i j, o h -> b o i j", qk_scores, self.sample_projection.weight
                )

            qk_scores = self.dropout(qk_scores)

            output_with_heads = qk_scores @ v

            output_without_heads = rearrange(output_with_heads, "b h t d -> b t (h d)")

            return self.out_proj(output_without_heads)

    def attention_logits(self, q, k, v):

        q, k, v = self.project_qkv(q, k, v)

        # Divide Q/K/V into heads
        q = rearrange(q, "b t (h d) -> b h t d", h=self.n_heads)
        k = rearrange(k, "b t (h d) -> b h t d", h=self.n_heads)
        v = rearrange(v, "b t (h d) -> b h t d", h=self.n_heads)

        qk_scores = q @ k.transpose(-1, -2)

        qk_scores *= self.scaling_factor

        # Apply mask if causal (must come before softmax)
        if self.causal:
            qk_scores.masked_fill_(self.mask, float("-inf"))

        return qk_scores  # (batch, head, seq_len, seq_len)

    def reset_parameters(self):
        # Default nn.Linear init is kaiming_uniform, which is fine
        self.q_proj.reset_parameters()
        self.k_proj.reset_parameters()
        self.v_proj.reset_parameters()
        scale_parameters(self.v_proj, self.beta)
        self.out_proj.reset_parameters()
        scale_parameters(self.out_proj, self.beta)

        if self.talking_heads:
            # Initialize close to identity
            nn.init.eye_(self.head_projection.weight)
            nn.init.eye_(self.sample_projection.weight)


class FeedforwardBlock(nn.Module):
    """
    ...
    """

    def __init__(
        self,
        input_features,
        ratio,
        output_features,
        activation=nn.ReLU,
        activation_kwargs=None,
        dropout=0.0,
        inner_dropout=None,
        outer_dropout=None,
        linear_module_up=nn.Linear,
        linear_module_down=nn.Linear,
        normformer=False,
        checkpoint=True,
        beta=1.0,
    ):
        super().__init__()

        self.checkpoint = checkpoint
        self.beta = beta
        self.xglu = activation.__name__.endswith("GLU")

        if activation_kwargs is not None:
            self.activation = activation(**activation_kwargs)
        else:
            self.activation = activation()

        self.inner_dropout = nn.Dropout(
            inner_dropout if inner_dropout is not None else dropout
        )
        self.outer_dropout = nn.Dropout(
            outer_dropout if outer_dropout is not None else dropout
        )

        self.max_features = (
            2 * int(ratio * output_features)
            if self.xglu
            else int(ratio * output_features)
        )

        self.linear_in = linear_module_up(input_features, self.max_features)
        self.linear_out = linear_module_down(
            int(ratio * output_features), output_features
        )

        self.process = nn.Sequential(
            *[
                self.linear_in,
                self.activation,
                (
                    nn.RMSNorm(int(ratio * output_features))
                    if normformer
                    else nn.Identity()
                ),
                self.inner_dropout,
                self.linear_out,
                (nn.RMSNorm(output_features) if normformer else nn.Identity()),
                self.outer_dropout,
            ]
        )

        self.recycling_enabled = False
        if hasattr(self.linear_in, "row_recycling_rate") and hasattr(
            self.linear_out, "column_recycling_rate"
        ):
            self.recycling_enabled = True
            self.master_recycling_rate = self.linear_in.row_recycling_rate
            self.linear_in.row_recycling_rate = 0.0
            self.linear_out.column_recycling_rate = 0.0
            if (
                hasattr(self.linear_in, "column_recycling_rate")
                and self.linear_in.column_recycling_rate > 0
            ) or (
                hasattr(self.linear_out, "row_recycling_rate")
                and self.linear_out.row_recycling_rate > 0
            ):
                raise NotImplementedError(
                    "At the moment this layer can only support recycling linear "
                    "layers if the in layer resets only rows and the out layer "
                    "resets only columns."
                )

        self.reset_parameters()

    def forward(self, x):

        # Recycle weights if using recycling linear layers
        if self.training and self.recycling_enabled:
            indices = self.linear_out.get_reset_indices(1)
            self.linear_in.reset_rows(indices, incoming_data=x)
            self.linear_out.reset_columns(indices)

        if self.checkpoint:
            processed = checkpoint(self.process, x, use_reentrant=False)
        else:
            processed = self.process(x)

        return processed

    def reset_parameters(self):
        # Iterate over the sequential block to reset parameters
        for module in self.process:
            if hasattr(module, "reset_parameters"):
                module.reset_parameters()

        scale_parameters(self.linear_in, self.beta)
        scale_parameters(self.linear_out, self.beta)


class EncoderBlock(nn.Module):
    """ """

    def __init__(
        self,
        seq_len,
        d_model,
        n_heads,
        relative_position_embedding=False,
        source_size=None,
        utility_tokens=0,
        talking_heads=False,
        mlp_ratio=4,
        activation: nn.Module = nn.ReLU,
        activation_kwargs: Optional[dict] = None,
        ff_linear_module_up=None,
        ff_linear_module_down=None,
        msa_scaling="d",
        ff_dropout=0.0,
        ff_inner_dropout=0.0,
        ff_outer_dropout=0.0,
        msa_dropout=0.0,
        identity_probability=0.0,
        causal=False,
        linear_module=nn.Linear,
        pre_norm=True,
        post_norm=False,
        normformer=False,
        checkpoint_ff=True,
        alpha=1.0,
        beta=1.0,
    ):
        """
        Args:
            msa_scaling: how should the attention logits be scaled? Can be "sqrtd"
                to mimic the original Attention is All You Need approach of
                dividing by the sqrt of the embedding Dimension or "d" per
                "Tensor Programs V...". Default "d"
        """

        super().__init__()

        if pre_norm and post_norm:
            raise ValueError("A transformer cannot be both prenorm and postnorm.")

        self.pre_norm = pre_norm
        self.post_norm = post_norm
        self.normformer = normformer

        self.alpha = alpha
        self.beta = beta

        self.drop_path = DropPath(drop_prob=identity_probability, scale_by_keep=True)

        if self.pre_norm:
            self.pre_attention_norm = nn.RMSNorm(d_model)
            self.pre_mlp_norm = nn.RMSNorm(d_model)

        if normformer:
            self.normformer_norm = nn.RMSNorm(d_model)

        if self.post_norm:
            self.input_norm = nn.RMSNorm(d_model)
            self.post_attention_norm = nn.RMSNorm(d_model)
            self.post_mlp_norm = nn.RMSNorm(d_model)

        if relative_position_embedding:
            max_freq = int(max(source_size) / 2)  # Suggested by Gemini!
            if d_model < 16:
                dim = d_model
            else:
                dim = 16
            self.rotary_embedding = RotaryEmbedding(
                dim=dim, freqs_for="pixel", max_freq=max_freq
            )
        else:
            self.rotary_embedding = None

        self.attn = MHAttention(  # Handles QKV projection
            d_model,
            n_heads,
            dropout=msa_dropout,
            causal=causal,
            seq_len=seq_len,
            linear_module=linear_module,
            rotary_embedding=self.rotary_embedding,
            source_size=source_size,
            utility_tokens=utility_tokens,
            talking_heads=talking_heads,
            scaling=msa_scaling,
            beta=beta,
        )

        # Submodule for the feedforward process
        self.ff = FeedforwardBlock(
            d_model,
            mlp_ratio,
            d_model,
            activation=activation,
            activation_kwargs=activation_kwargs,
            dropout=ff_dropout,
            inner_dropout=ff_inner_dropout,
            outer_dropout=ff_outer_dropout,
            linear_module_up=(
                ff_linear_module_up
                if ff_linear_module_up is not None
                else linear_module
            ),
            linear_module_down=(
                ff_linear_module_down
                if ff_linear_module_down is not None
                else linear_module
            ),
            normformer=normformer,
            checkpoint=checkpoint_ff,
            beta=beta,
        )

        self.reset_parameters()

    @property
    def _kv_distance(self) -> float:
        return self.attn._kv_distance

    def forward(self, x):

        if self.pre_norm:
            process_x = self.pre_attention_norm(x)
        else:
            process_x = x

        processed = self.attn(process_x, process_x, process_x)

        if self.normformer:
            processed = self.normformer_norm(processed)

        processed = self.drop_path(processed)

        x = self.alpha * x + processed

        if self.post_norm:
            x = self.post_attention_norm(x)
            process_x = x
        elif self.pre_norm:
            process_x = self.pre_mlp_norm(x)
        else:
            process_x = x

        processed = self.drop_path(self.ff(process_x))

        x = self.alpha * x + processed

        if self.post_norm:
            x = self.post_mlp_norm(x)

        return x

    def attention_logits(self, x):
        """
        Give back the attention scores used in this layer.
        Needs to match what the model actually sees during forward()
        by applying the correct normalisations.
        """
        if self.pre_norm:
            x = self.pre_attention_norm(x)
        elif self.post_norm:
            x = self.input_norm(x)

        return self.attn.attention_logits(x, x, x)

    def reset_parameters(self):
        if self.pre_norm:
            self.pre_attention_norm.reset_parameters()
            self.pre_mlp_norm.reset_parameters()

        if self.post_norm:
            self.post_attention_norm.reset_parameters()
            self.post_mlp_norm.reset_parameters()

        if self.normformer:
            self.normformer_norm.reset_parameters()

        self.attn.reset_parameters()
        self.ff.reset_parameters()


class TransformerEncoder(nn.Module):
    """
    This assumes we already get a sequence of embeddings (e.g. word or image
        patch embeddings).
    """

    def __init__(
        self,
        seq_len,
        d_model,
        n_layers,
        n_heads,
        absolute_position_embedding=True,
        relative_position_embedding=False,
        source_size=None,
        mlp_ratio=4,
        activation: nn.Module = nn.ReLU,
        activation_kwargs: Optional[dict] = None,
        ff_linear_module_up=None,
        ff_linear_module_down=None,
        ff_dropout=0.0,
        ff_inner_dropout=0.0,
        ff_outer_dropout=0.0,
        msa_dropout=0.0,
        stochastic_depth=0.0,
        causal=False,
        linear_module=nn.Linear,
        utility_tokens=0,
        talking_heads=False,
        return_utility_tokens=False,
        pre_norm=True,
        post_norm=False,
        normformer=False,
        msa_scaling="d",
        checkpoint_ff=True,
        alpha=1.0,
        beta=1.0,
    ):
        """
        Args:
            msa_scaling: how should the attention logits be scaled? Can be "sqrtd"
                to mimic the original Attention is All You Need approach of
                dividing by the sqrt of the embedding Dimension or "d" per
                "Tensor Programs V...". Default "d"
        """

        if relative_position_embedding and (source_size is None):
            raise ValueError(
                "`source_size` for TransformerEncoder cannot be None if"
                " `relative_position_embedding` is True"
            )

        if absolute_position_embedding and (seq_len is None):
            raise ValueError(
                "`seq_len` for TransformerEncoder cannot be None if"
                " `absolute_position_embedding` is True"
            )

        super().__init__()

        if FLASH_ATTN and talking_heads:
            warnings.warn(
                "Using talking heads currently prevents using flash attention.",
                stacklevel=2,
            )

        self.seq_len = seq_len
        self.n_heads = n_heads
        self._utility_tokens = utility_tokens
        self.return_utility_tokens = return_utility_tokens

        # Initialise utility tokens with normal init, like usual Pytorch embeddings
        if self._utility_tokens:
            self._utility_token_embedding = nn.Parameter(
                torch.empty(self._utility_tokens, d_model)
            )
            nn.init.normal_(self._utility_token_embedding, mean=0.0, std=1.0)
        else:
            self._utility_token_embedding = None

        if self._utility_tokens and (self.seq_len is not None):
            self.full_sequence_length = self.seq_len + self._utility_tokens
        else:
            self.full_sequence_length = self.seq_len

        self.d_model = d_model

        if absolute_position_embedding:
            self.absolute_position_embedding = nn.Embedding(
                self.full_sequence_length, d_model
            )
        else:
            self.absolute_position_embedding = None

        self.mlp_dropout = ff_dropout
        self.msa_dropout = msa_dropout
        self.stochastic_depth = stochastic_depth

        assert isinstance(n_layers, int)

        if n_layers == 1:
            self.stochastic_depth_probabilities = [0.0]
        else:
            step_size = self.stochastic_depth / (n_layers - 1)
            self.stochastic_depth_probabilities = [
                i * step_size for i in range(n_layers)
            ]

        self.blocks = nn.ModuleList(
            [
                EncoderBlock(
                    self.full_sequence_length,
                    d_model,
                    n_heads,
                    relative_position_embedding=relative_position_embedding,
                    source_size=source_size,
                    utility_tokens=utility_tokens,
                    talking_heads=talking_heads,
                    mlp_ratio=mlp_ratio,
                    activation=activation,
                    activation_kwargs=activation_kwargs,
                    ff_linear_module_up=ff_linear_module_up,
                    ff_linear_module_down=ff_linear_module_down,
                    msa_scaling=msa_scaling,
                    ff_dropout=ff_dropout,
                    ff_inner_dropout=ff_inner_dropout,
                    ff_outer_dropout=ff_outer_dropout,
                    msa_dropout=msa_dropout,
                    identity_probability=self.stochastic_depth_probabilities[i],
                    causal=causal,
                    linear_module=linear_module,
                    pre_norm=pre_norm,
                    post_norm=post_norm,
                    normformer=normformer,
                    checkpoint_ff=checkpoint_ff,
                    alpha=alpha,
                    beta=beta,
                )
                for i in range(n_layers)
            ]
        )

        self.reset_parameters()

    @property
    def _kv_distances(self) -> float:
        return ",".join([str(block._kv_distance) for block in self.blocks])

    def preprocess(self, x):
        if self._utility_tokens:
            x = torch.cat(
                [self._utility_token_embedding.expand(x.size(0), -1, -1), x], dim=1
            )
        else:
            x = x

        if self.absolute_position_embedding is not None:
            position_embedding = self.absolute_position_embedding(
                torch.arange(
                    0, self.full_sequence_length, dtype=torch.long, device=x.device
                ).unsqueeze(
                    0
                )  # to shape (1, seq_len) to broadcast over batch
            )
            x += position_embedding

        return x

    def forward(self, x):

        x = self.preprocess(x)

        for block in self.blocks:
            x = block(x)

        if self._utility_tokens and not self.return_utility_tokens:
            return x[:, self._utility_tokens :, :]
        else:
            return x

    def attention_logits(self, x):

        x = self.preprocess(x)

        layer_scores = []

        for block in self.blocks:
            # Get attention scores with shape (batch, 1, head, seq_len, seq_len)
            layer_attention_logits = block.attention_logits(x).unsqueeze(1)
            layer_scores.append(layer_attention_logits)
            x = block(x)

        return torch.cat(layer_scores, dim=1)  # (batch, layer, head, seq_len, seq_len)

    def reset_parameters(self):
        if self._utility_token_embedding is not None:
            nn.init.normal_(self._utility_token_embedding, mean=0.0, std=1.0)

        if self.absolute_position_embedding is not None:
            self.absolute_position_embedding.reset_parameters()

        for block in self.blocks:
            block.reset_parameters()



import torch
from torch import nn
from torch.nn import functional as F


class SigmaReparamTensor(nn.Module):
    """
    Inspired by Apple's Spectral Normed Linear Layers
        (https://github.com/apple/ml-sigma-reparam)
    """

    def __init__(self, init_tensor: torch.Tensor):
        assert init_tensor.ndim == 2

        super().__init__()

        self.sigma_reparam_tensor = nn.Parameter(init_tensor, requires_grad=True)

        with torch.no_grad():
            _, sigma, v_transpose = torch.linalg.svd(
                self.sigma_reparam_tensor, full_matrices=False
            )

        self.register_buffer("approx_spectral_norm", sigma[:1])
        self.register_buffer("right_singular", v_transpose[0])
        self.sigma_reparam_scale = nn.Parameter(
            self.approx_spectral_norm.clone().detach(), requires_grad=True
        )

    def power_iteration(self):
        with torch.no_grad():
            approx_right_singular_transpose = self.sigma_reparam_tensor.mv(
                self.right_singular
            )
            approx_right_singular_transpose = F.normalize(
                approx_right_singular_transpose, dim=0
            )
            updated_right_singular = self.sigma_reparam_tensor.T.mv(
                approx_right_singular_transpose
            )
            updated_right_singular = F.normalize(updated_right_singular, dim=0)
            self.right_singular.data.copy_(updated_right_singular)
            rayleigh_quotient = torch.einsum(
                "m,mn,n->",
                approx_right_singular_transpose,
                self.sigma_reparam_tensor,
                updated_right_singular,
            )
            self.approx_spectral_norm.data.copy_(rayleigh_quotient)

    def forward(self):
        if self.training:
            self.power_iteration()
        return self.sigma_reparam_scale * (
            self.sigma_reparam_tensor / self.approx_spectral_norm
        )


class AnchoredReparamTensor(nn.Module):
    """
    Reparameterises a tensor by decoupling its magnitude and direction.

    The direction is represented by a learnable weight tensor, normalised by the
    Rayleigh quotient with respect to its initial dominant right-singular vector.
    The magnitude is a separate learnable scalar.

    The reparameterization is:

        W_reparam = scale * (W / norm)

    where the norm is the Rayleigh quotient uᵀWv₀, with v₀ being the dominant
    right-singular vector of the initial tensor and u = normalize(Wv₀).
    """

    def __init__(self, init_tensor: torch.Tensor):
        assert init_tensor.ndim == 2

        super().__init__()

        self.weight = nn.Parameter(init_tensor, requires_grad=True)

        with torch.no_grad():
            _, sigma, v_transpose = torch.linalg.svd(self.weight, full_matrices=False)

        self.register_buffer("rayleigh_norm", sigma[:1])
        self.register_buffer("initial_right_singular", v_transpose[0])
        self.nondecay_scale = nn.Parameter(
            sigma[:1].clone().detach(), requires_grad=True
        )

    def _update_rayleigh_norm(self):
        with torch.no_grad():
            product = self.weight.mv(self.initial_right_singular)
            normed_product = F.normalize(product, dim=0)
            rayleigh_norm = torch.einsum(
                "m,mn,n->",
                normed_product,
                self.weight,
                self.initial_right_singular,
            )
            self.rayleigh_norm.data.copy_(rayleigh_norm)

    def forward(self):
        if self.training:
            self._update_rayleigh_norm()
        return self.nondecay_scale * (self.weight / (self.rayleigh_norm + 1e-6))


class NormReparamTensor(nn.Module):
    """
    Reparameterise a tensor as a normalised tensor of weights multiplied by a
        learnable scaling factor.
    """

    def __init__(self, init_tensor: torch.Tensor):
        assert init_tensor.ndim == 2, "Input tensor must be a 2D matrix."
        super().__init__()

        # Use the gradboard convention of calling something nondecay_* if we should
        # exclude it from weight decay
        self.weight = nn.Parameter(init_tensor.clone(), requires_grad=True)
        self.nondecay_scale = nn.Parameter(
            torch.linalg.norm(self.weight).clone().detach(), requires_grad=True
        )

    def forward(self) -> torch.Tensor:
        norm = torch.linalg.norm(self.weight)
        return self.nondecay_scale * (self.weight / (norm + 1e-6))



import math
from typing import Optional

from .transformer import TransformerEncoder, FeedforwardBlock
from .cnn import SpaceToDepth, calculate_output_spatial_size, spatial_tuple
from .activation import ReLU, SquaredReLU, GELU, SwiGLU
from .utils import PadTensor

from einops import einsum
from einops.layers.torch import Rearrange

import torch
import torch.nn as nn


class GetCLSToken(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        return x[:, 0, :]


class SequencePool(nn.Module):
    def __init__(self, d_model, linear_module=nn.Linear):
        super().__init__()
        self.attention = nn.Sequential(
            *[
                linear_module(d_model, 1),
                Rearrange("batch seq 1 -> batch seq"),
                nn.Softmax(dim=-1),
            ]
        )

        self.reset_parameters()

    def forward(self, x):
        weights = self.attention(x)
        return einsum(weights, x, "batch seq, batch seq d_model -> batch d_model")

    def attention_scores(self, x):
        return self.attention(x)

    def reset_parameters(self):
        # Iterate over modules in the sequential block
        for module in self.attention:
            if hasattr(module, "reset_parameters"):
                module.reset_parameters()


class ClassificationHead(nn.Module):
    """
    A general classification head for a ViT
    """

    def __init__(
        self,
        d_model,
        n_classes,
        logit_projection_layer=nn.Linear,
        batch_norm_logits=True,
    ):
        super().__init__()
        self.d_model = d_model
        self.summarize = GetCLSToken()

        if d_model == n_classes:
            # No need to project
            self.projection = nn.Identity()
        else:
            self.projection = logit_projection_layer(d_model, n_classes)

        if batch_norm_logits:
            self.batch_norm = nn.BatchNorm1d(n_classes, affine=False)
        else:
            self.batch_norm = nn.Identity()

        self.classification_process = nn.Sequential(
            *[
                self.summarize,
                self.projection,
                self.batch_norm,
            ]
        )

        self.reset_parameters()

    def forward(self, x):
        return self.classification_process(x)

    def reset_parameters(self):
        for module in self.classification_process:
            if hasattr(module, "reset_parameters"):
                module.reset_parameters()


class SequencePoolClassificationHead(ClassificationHead):
    """
    As described in [Hasani et al. (2021) *''Escaping the Big Data Paradigm with
        Compact Transformers''*](https://arxiv.org/abs/2104.05704). It can be viewed
        as a generalisation of average pooling.
    """

    def __init__(
        self,
        d_model,
        n_classes,
        logit_projection_layer=nn.Linear,
        batch_norm_logits=True,
    ):
        super().__init__(
            d_model,
            n_classes,
            logit_projection_layer=logit_projection_layer,
            batch_norm_logits=batch_norm_logits,
        )

        self.summarize = SequencePool(d_model, logit_projection_layer)
        # Rebuild the classification process with the correct summary module:
        self.classification_process = nn.Sequential(
            *[
                self.summarize,
                self.projection,
                self.batch_norm,
            ]
        )

        self.reset_parameters()


class ViTEncoder(nn.Module):
    """
    Based on the Compact Convolutional Transformer (CCT) of [Hasani et al. (2021)
        *''Escaping the Big Data Paradigm with Compact Transformers''*](
        https://arxiv.org/abs/2104.05704). It's basically a convolutional neural
        network leading into a transformer encoder. To make it like the full CCT
        we would finish it of with a sequence pooling layer but we won't always
        want to do that.
    """

    def __init__(
        self,
        input_size=(32, 32),
        in_channels=3,
        initial_batch_norm=True,
        cnn=True,
        cnn_out_channels=16,
        cnn_kernel_size=3,
        cnn_kernel_stride=1,
        cnn_padding="same",
        cnn_kernel_dilation=1,
        cnn_kernel_groups=1,
        cnn_activation: nn.Module = ReLU,
        cnn_activation_kwargs: Optional[dict] = None,
        cnn_dropout=0.0,
        pooling_type="concat",  # max, average or concat
        pooling_kernel_size=3,
        pooling_kernel_stride=2,
        pooling_padding=1,
        transformer_feedforward_first=True,
        transformer_initial_ff_linear_module_up=None,
        transformer_initial_ff_linear_module_down=None,
        transformer_initial_ff_dropout=None,
        transformer_initial_ff_inner_dropout=None,
        transformer_initial_ff_outer_dropout=None,
        transformer_pre_norm=True,
        transformer_normformer=False,
        transformer_post_norm=False,
        transformer_absolute_position_embedding=False,
        transformer_relative_position_embedding=True,
        transformer_embedding_size=256,
        transformer_layers=7,
        transformer_heads=4,
        transformer_mlp_ratio=2,
        transformer_utility_tokens=0,
        transformer_talking_heads=False,
        transformer_return_utility_tokens=False,
        transformer_activation: nn.Module = SquaredReLU,
        transformer_activation_kwargs: Optional[dict] = None,
        transformer_ff_linear_module_up=None,
        transformer_ff_linear_module_down=None,
        transformer_msa_scaling="d",
        transformer_ff_dropout=0.0,
        transformer_ff_inner_dropout=0.0,
        transformer_ff_outer_dropout=0.0,
        transformer_msa_dropout=0.1,
        transformer_stochastic_depth=0.1,
        transformer_checkpoint_ff=True,
        linear_module=nn.Linear,
        alpha=1.0,
        beta=1.0,
    ):
        super().__init__()

        self.alpha = alpha
        self.beta = beta

        if cnn_activation_kwargs is not None:
            self.cnn_activation = cnn_activation(**cnn_activation_kwargs)
        else:
            self.cnn_activation = cnn_activation()

        if transformer_activation_kwargs is not None:
            self.transformer_activation = transformer_activation(
                **transformer_activation_kwargs
            )
        else:
            self.transformer_activation = transformer_activation()

        self.input_size = input_size
        self.spatial_dimensions = len(self.input_size)

        if self.spatial_dimensions == 1:
            maxpoolxd = nn.MaxPool1d
            avgpoolxd = nn.AvgPool1d
            convxd = nn.Conv1d
            batchnormxd = nn.BatchNorm1d
            spatial_dim_names = "D1"
        elif self.spatial_dimensions == 2:
            maxpoolxd = nn.MaxPool2d
            avgpoolxd = nn.AvgPool2d
            convxd = nn.Conv2d
            batchnormxd = nn.BatchNorm2d
            spatial_dim_names = "D1 D2"
        elif self.spatial_dimensions == 3:
            maxpoolxd = nn.MaxPool3d
            avgpoolxd = nn.AvgPool3d
            convxd = nn.Conv3d
            batchnormxd = nn.BatchNorm3d
            spatial_dim_names = "D1 D2 D3"
        else:
            raise NotImplementedError(
                "`input_size` must be a tuple of length 1, 2, or 3."
            )

        if cnn:
            # This block rhymes:
            if cnn_activation.__name__.endswith("GLU"):
                cnn_out_channels *= 2
            cnn_output_size = calculate_output_spatial_size(
                input_size,
                kernel_size=cnn_kernel_size,
                stride=cnn_kernel_stride,
                padding=cnn_padding,
                dilation=cnn_kernel_dilation,
            )
            self.cnn = convxd(
                in_channels,
                cnn_out_channels,
                cnn_kernel_size,
                stride=cnn_kernel_stride,
                padding=cnn_padding,
                dilation=cnn_kernel_dilation,
                groups=cnn_kernel_groups,
                bias=True,
                padding_mode="zeros",
            )
            cnn_activation_out_channels = cnn_out_channels
            self.activate_and_dropout = nn.Sequential(
                *[
                    Rearrange(  # rearrange in case we're using XGLU activation
                        f"N C {spatial_dim_names} -> N {spatial_dim_names} C"
                    ),
                    self.cnn_activation,
                    Rearrange(f"N {spatial_dim_names} C -> N C {spatial_dim_names}"),
                    nn.Dropout(cnn_dropout),
                    batchnormxd(cnn_activation_out_channels),
                ]
            )
        else:
            self.cnn = nn.Identity()
            self.activate_and_dropout = nn.Identity()
            cnn_output_size = input_size
            cnn_out_channels = in_channels
            cnn_activation_out_channels = in_channels

        pooling_kernel_voxels = math.prod(
            spatial_tuple(pooling_kernel_size, self.spatial_dimensions)
        )

        pooling_output_size = (
            cnn_output_size
            if pooling_type is None
            else calculate_output_spatial_size(
                cnn_output_size,
                kernel_size=pooling_kernel_size,
                stride=pooling_kernel_stride,
                padding=pooling_padding,
                dilation=1,
            )
        )

        if pooling_type is None:
            pooling_out_channels = cnn_activation_out_channels
            self.pool = nn.Identity()

        elif pooling_type == "max":
            pooling_out_channels = cnn_activation_out_channels
            self.pool = maxpoolxd(
                pooling_kernel_size,
                stride=pooling_kernel_stride,
                padding=pooling_padding,
            )
        elif pooling_type == "average":
            pooling_out_channels = cnn_activation_out_channels
            self.pool = avgpoolxd(
                pooling_kernel_size,
                stride=pooling_kernel_stride,
                padding=pooling_padding,
            )
        elif pooling_type == "concat":
            pooling_out_channels = pooling_kernel_voxels * cnn_activation_out_channels
            self.pool = SpaceToDepth(
                pooling_kernel_size,
                stride=pooling_kernel_stride,
                padding=pooling_padding,
                spatial_dimensions=self.spatial_dimensions,
            )
        else:
            raise NotImplementedError(
                "Pooling type must be max, average, concat or None"
            )

        self.pooling_channels_padding = PadTensor(
            (0, max(0, transformer_embedding_size - pooling_out_channels))
        )

        self.sequence_length = math.prod(pooling_output_size)  # One token per voxel

        if transformer_layers > 0:
            self.transformer = TransformerEncoder(
                self.sequence_length,
                transformer_embedding_size,
                transformer_layers,
                transformer_heads,
                absolute_position_embedding=transformer_absolute_position_embedding,
                relative_position_embedding=transformer_relative_position_embedding,
                source_size=pooling_output_size,
                mlp_ratio=transformer_mlp_ratio,
                activation=transformer_activation,
                activation_kwargs=transformer_activation_kwargs,
                ff_linear_module_up=transformer_ff_linear_module_up,
                ff_linear_module_down=transformer_ff_linear_module_down,
                msa_scaling=transformer_msa_scaling,
                ff_dropout=transformer_ff_dropout,
                ff_inner_dropout=transformer_ff_inner_dropout,
                ff_outer_dropout=transformer_ff_outer_dropout,
                msa_dropout=transformer_msa_dropout,
                stochastic_depth=transformer_stochastic_depth,
                causal=False,
                linear_module=linear_module,
                utility_tokens=transformer_utility_tokens,
                talking_heads=transformer_talking_heads,
                return_utility_tokens=transformer_return_utility_tokens,
                pre_norm=transformer_pre_norm,
                normformer=transformer_normformer,
                post_norm=transformer_post_norm,
                checkpoint_ff=transformer_checkpoint_ff,
                alpha=self.alpha,
                beta=self.beta,
            )
        else:
            self.transformer = nn.Identity()

        if transformer_feedforward_first:
            self.initial_ff = FeedforwardBlock(
                max(transformer_embedding_size, pooling_out_channels),
                transformer_mlp_ratio,
                transformer_embedding_size,
                activation=transformer_activation,
                activation_kwargs=transformer_activation_kwargs,
                dropout=(
                    # First truthy assigned value
                    transformer_initial_ff_dropout
                    if transformer_initial_ff_dropout is not None
                    else transformer_ff_dropout
                ),
                inner_dropout=(
                    # First truthy assigned value
                    transformer_initial_ff_inner_dropout
                    if transformer_initial_ff_inner_dropout is not None
                    else transformer_ff_inner_dropout
                ),
                outer_dropout=(
                    # First truthy assigned value
                    transformer_initial_ff_outer_dropout
                    if transformer_initial_ff_outer_dropout is not None
                    else transformer_ff_outer_dropout
                ),
                linear_module_up=(
                    # First truthy assigned value
                    transformer_initial_ff_linear_module_up
                    or transformer_ff_linear_module_up
                    or linear_module
                ),
                linear_module_down=(
                    # First truthy assigned value
                    transformer_initial_ff_linear_module_down
                    or transformer_ff_linear_module_down
                    or linear_module
                ),
                normformer=transformer_normformer,
                checkpoint=transformer_checkpoint_ff,
                beta=self.beta,
            )
            self.layer_norm = nn.RMSNorm(transformer_embedding_size)
        else:
            self.initial_ff = None

        self.preprocess = nn.Sequential(
            *[
                batchnormxd(in_channels) if initial_batch_norm else nn.Identity(),
                self.cnn,
                self.activate_and_dropout,
                self.pool,
                Rearrange(  # for transformer
                    f"N C {spatial_dim_names} -> N ({spatial_dim_names}) C"
                ),
                self.pooling_channels_padding,
                nn.RMSNorm(transformer_embedding_size),
            ]
        )

        self.reset_parameters()

    def forward(self, x):
        x = self.preprocess(x)
        if self.initial_ff is not None:
            x = self.layer_norm(x + self.initial_ff(x))
        return self.transformer(x)

    def attention_logits(self, x):
        x = self.preprocess(x)
        x = x + self.initial_ff(x)
        return self.transformer.attention_logits(x)

    def reset_parameters(self):
        for module in self.preprocess:
            if hasattr(module, "reset_parameters"):
                module.reset_parameters()
        self.initial_ff.reset_parameters()
        self.transformer.reset_parameters()


class ViT(nn.Module):
    """
    ...
    """

    def __init__(
        self,
        input_size=(32, 32),
        image_classes=100,
        in_channels=3,
        initial_batch_norm=True,
        cnn=True,
        cnn_out_channels=16,
        cnn_kernel_size=3,
        cnn_kernel_stride=1,
        cnn_padding="same",
        cnn_kernel_dilation=1,
        cnn_kernel_groups=1,
        cnn_activation: nn.Module = ReLU,
        cnn_activation_kwargs: Optional[dict] = None,
        cnn_dropout=0.0,
        pooling_type="concat",  # max, average or concat
        pooling_kernel_size=3,
        pooling_kernel_stride=2,
        pooling_padding=1,
        transformer_feedforward_first=True,
        transformer_initial_ff_linear_module_up=None,
        transformer_initial_ff_linear_module_down=None,
        transformer_initial_ff_dropout=None,
        transformer_initial_ff_inner_dropout=None,
        transformer_initial_ff_outer_dropout=None,
        transformer_pre_norm=True,
        transformer_normformer=False,
        transformer_post_norm=False,
        transformer_absolute_position_embedding=False,
        transformer_relative_position_embedding=True,
        transformer_embedding_size=256,
        transformer_layers=7,
        transformer_heads=4,
        transformer_mlp_ratio=2,
        transformer_utility_tokens=0,
        transformer_talking_heads=False,
        transformer_return_utility_tokens=False,
        transformer_activation: nn.Module = SquaredReLU,
        transformer_activation_kwargs: Optional[dict] = None,
        transformer_ff_linear_module_up=None,
        transformer_ff_linear_module_down=None,
        transformer_msa_scaling="d",
        transformer_ff_dropout=0.0,
        transformer_ff_inner_dropout=0.0,
        transformer_ff_outer_dropout=0.0,
        transformer_msa_dropout=0.1,
        transformer_stochastic_depth=0.1,
        transformer_checkpoint_ff=True,
        head=SequencePoolClassificationHead,
        batch_norm_logits=True,
        logit_projection_layer=nn.Linear,
        linear_module=nn.Linear,
    ):

        super().__init__()

        if isinstance(cnn_activation, str):
            cnn_activation = {
                "ReLU": ReLU,
                "SquaredReLU": SquaredReLU,
                "GELU": GELU,
                "SwiGLU": SwiGLU,
            }[cnn_activation]

        if isinstance(transformer_activation, str):
            transformer_activation = {
                "ReLU": ReLU,
                "SquaredReLU": SquaredReLU,
                "GELU": GELU,
                "SwiGLU": SwiGLU,
            }[transformer_activation]

        self.alpha = (2 * transformer_layers) ** 0.25
        self.beta = (8 * transformer_layers) ** 0.25

        self.encoder = ViTEncoder(
            input_size=input_size,
            initial_batch_norm=initial_batch_norm,
            in_channels=in_channels,
            cnn=cnn,
            cnn_out_channels=cnn_out_channels,
            cnn_kernel_size=cnn_kernel_size,
            cnn_kernel_stride=cnn_kernel_stride,
            cnn_padding=cnn_padding,
            cnn_kernel_dilation=cnn_kernel_dilation,
            cnn_kernel_groups=cnn_kernel_groups,
            cnn_activation=cnn_activation,
            cnn_activation_kwargs=cnn_activation_kwargs,
            cnn_dropout=cnn_dropout,
            pooling_type=pooling_type,
            pooling_kernel_size=pooling_kernel_size,
            pooling_kernel_stride=pooling_kernel_stride,
            pooling_padding=pooling_padding,
            transformer_feedforward_first=transformer_feedforward_first,
            transformer_initial_ff_linear_module_up=transformer_initial_ff_linear_module_up,
            transformer_initial_ff_linear_module_down=transformer_initial_ff_linear_module_down,
            transformer_initial_ff_dropout=transformer_initial_ff_dropout,
            transformer_initial_ff_inner_dropout=transformer_initial_ff_inner_dropout,
            transformer_initial_ff_outer_dropout=transformer_initial_ff_outer_dropout,
            transformer_pre_norm=transformer_pre_norm,
            transformer_normformer=transformer_normformer,
            transformer_post_norm=transformer_post_norm,
            transformer_absolute_position_embedding=transformer_absolute_position_embedding,
            transformer_relative_position_embedding=transformer_relative_position_embedding,
            transformer_embedding_size=transformer_embedding_size,
            transformer_layers=transformer_layers,
            transformer_heads=transformer_heads,
            transformer_mlp_ratio=transformer_mlp_ratio,
            transformer_utility_tokens=transformer_utility_tokens,
            transformer_talking_heads=transformer_talking_heads,
            transformer_return_utility_tokens=transformer_return_utility_tokens,
            transformer_activation=transformer_activation,
            transformer_activation_kwargs=transformer_activation_kwargs,
            transformer_ff_linear_module_up=transformer_ff_linear_module_up,
            transformer_ff_linear_module_down=transformer_ff_linear_module_down,
            transformer_msa_scaling=transformer_msa_scaling,
            transformer_ff_dropout=transformer_ff_dropout,
            transformer_ff_inner_dropout=transformer_ff_inner_dropout,
            transformer_ff_outer_dropout=transformer_ff_outer_dropout,
            transformer_msa_dropout=transformer_msa_dropout,
            transformer_stochastic_depth=transformer_stochastic_depth,
            transformer_checkpoint_ff=transformer_checkpoint_ff,
            linear_module=linear_module,
            alpha=self.alpha,
            beta=self.beta,
        )

        self.pool = head(
            transformer_embedding_size,
            image_classes,
            logit_projection_layer=logit_projection_layer,
            batch_norm_logits=batch_norm_logits,
        )

        self.reset_parameters()

    @property
    def sequence_length(self):
        return self.encoder.sequence_length

    def forward(self, x):
        return self.pool(self.encoder(x))

    def attention_logits(self, x):
        return self.encoder.attention_logits(x)

    def pool_attention(self, x):
        if hasattr(self.pool.summarize, "attention"):
            return self.pool.summarize.attention(self.encoder(x))
        else:
            raise NotImplementedError(
                "`pool_attention` is currently only implemented where"
                " head class is SequencePoolClassificationHead"
            )

    def head_to_utility_token_attention_logits(self, x):
        all_attention = self.attention_logits(x)
        batch_averages = torch.mean(all_attention, dim=0, keepdim=False)
        sequence_averages = torch.mean(batch_averages, dim=-1, keepdim=False)
        n_utility_tokens = self.encoder.encoder[-1]._utility_tokens
        return sequence_averages[
            :, :, :n_utility_tokens
        ]  # (layer, head, utility_tokens)

    def reset_parameters(self):
        self.encoder.reset_parameters()
        self.pool.reset_parameters()



import torch
import torch.nn as nn
import torch.nn.functional as F
import math
from typing import Union

from einops.layers.torch import Rearrange


def spatial_tuple(size: Union[int, tuple], spatial_dimensions):
    """
    Converts an integer x to `tuple([x] * spatial_dimensions)`.
    Performs no operation (i.e. the identity operation) on tuples of length `spatial_dimensions`.
    Otherwise
    """
    if isinstance(size, int):
        return tuple([size] * spatial_dimensions)
    elif isinstance(size, tuple) and (len(size) == spatial_dimensions):
        return size
    else:
        raise ValueError(
            f"For {spatial_dimensions} spatial dimensions, `size` must be "
            f"an integer or a tuple of length {spatial_dimensions}."
        )


def padding_tensor(padding: tuple):
    """
    Converts a tuple of ints (x, y, z) into a tuple of 2-tuples,
        like ((x, x), (y, y), (z, z)).

    Performs no operation (i.e. the identity operation) on a tuple of 2-tuples.

    Otherwise raises an error.
    """
    if all(isinstance(x, int) for x in padding):
        return tuple([tuple([p] * 2) for p in padding])
    elif (
        all(isinstance(p, tuple) for p in padding)
        and all(len(p) == 2 for p in padding)
        and all(all(isinstance(x, int) for x in p) for p in padding)
    ):
        return padding
    else:
        raise ValueError(
            "Padding must be a tuple of ints of a tuple of 2-tuples of ints. "
            f"It was {padding}."
        )


def kd_unfold(t: torch.Tensor, kernel_size=1, stride=1, padding=0, k=2):
    """
    Unfold operation with k spatial dimensions.
    Does not support dilation.
    Only supports equal padding at top and bottom.
    """
    if len(t.size()[2:]) != k:
        raise ValueError(
            f"Input tensor size should be (N, channels, spatial dims...), so "
            f"for k = {k}, t.size() should be a tuple of length {k + 2}."
        )

    N, C = t.size(0), t.size(1)

    kernel_size = spatial_tuple(kernel_size, k)
    stride = spatial_tuple(stride, k)
    padding = padding_tensor(spatial_tuple(padding, k))

    output = t
    output = F.pad(output, sum(reversed(padding), ()))  # i.e. the empty tuple

    for i, _ in enumerate(kernel_size):
        output = output.unfold(i + 2, kernel_size[i], stride[i])

    permutation = [0, 1] + [i + k + 2 for i in range(k)] + [i + 2 for i in range(k)]

    return output.permute(*permutation).reshape(N, math.prod(kernel_size) * C, -1)


def calculate_output_spatial_size(
    input_spatial_size, kernel_size=1, stride=1, padding=0, dilation=0
):
    """
    Calculate the output size for the spatial dimensions of a convolutional operation
    """
    stride = spatial_tuple(stride, len(input_spatial_size))

    # Handle padding keywords that are sometimes used
    if padding == "same":
        output_size = ()
        for i, in_length in enumerate(input_spatial_size):
            output_size += (math.ceil(in_length / stride[i]),)
        return output_size
    elif padding == "valid":
        padding = 0

    kernel_size = spatial_tuple(kernel_size, len(input_spatial_size))
    padding = spatial_tuple(padding, len(input_spatial_size))
    dilation = spatial_tuple(dilation, len(input_spatial_size))

    output_size = ()

    for i, in_length in enumerate(input_spatial_size):
        output_size += (
            math.floor(
                (in_length + 2 * padding[i] - dilation[i] * (kernel_size[i] - 1) - 1)
                / stride[i]
                + 1
            ),
        )
    return output_size


class SpaceToDepth(nn.Module):
    """
    An operation that extracts patches from an image-like tensor and stacks
        them channel-wise.
    """

    def __init__(self, kernel_size, stride=1, padding=0, spatial_dimensions=2):
        """
        Input shape should be in order (channels, spatial dims...),
            e.g. (channels, height, width)
        """

        super().__init__()

        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.spatial_dimensions = spatial_dimensions

    def forward(self, x):

        N, C, *input_spatial_size = x.size()

        patches = kd_unfold(
            x,
            kernel_size=self.kernel_size,
            stride=self.stride,
            padding=self.padding,
            k=self.spatial_dimensions,
        )

        output_spatial_size = calculate_output_spatial_size(
            input_spatial_size=input_spatial_size,
            kernel_size=self.kernel_size,
            stride=self.stride,
            padding=self.padding,
            dilation=1,  # kd_unfold doesn't support dilation
        )

        output_channels = C * math.prod(
            spatial_tuple(self.kernel_size, self.spatial_dimensions)
        )

        return patches.view(N, output_channels, *output_spatial_size)
